#version 440 core

// ==================================================
// Código 'Fragment Shader'
// Este código (paralelizado) corre en la GPU para cada uno de los pixeles de una textura.
// Este código está escrito en lenguaje 'GLSL' (OpenGL Shading Language) y es muy similar al lenguaje 'C'.
// El código lo compila una API del propio OpenGL antes de pasarlo a la GPU (no lo compila MS Visual C++)
// ==================================================

// Variables de salida, hacia el 'pipeline' del render
out vec4 FragColor;

// Variables de entrada desde el código de 'Fragment Shader'
in vec4 vertexColor;
in vec2 vertexUv;
in vec3 vertexNorms;
in vec3 fragPos;
in float visibility; 

// Declaración y definición de la estructura de datos que definen la luz de la linterna del jugador
struct Spotlight
{
	vec3 position;
	vec3 direction;
	float cutOff;
	float outerCutOff;

	vec3 diffuse;
	vec3 specular;
	
	float constant;
	float linear;
	float quadratic;
};

// Variables de entrada. Se las pasa el código del programa de la CPU en cada 'frame'.
uniform sampler2D meshTexture;
uniform vec3 lightPos; 
uniform vec3 viewPos;
uniform bool damaged;	// 'damaged' llega con valor 'true' desde el programa de la CPU cuando un enemigo es alcanzado por un disparo del jugador
uniform vec3 colorize_model;	// 'colorize_model' llega desde la CPU con el color (RGB) a usar para colorear pixeles 
uniform bool nightFog;	// 'nightFog' llega con valor 'true' desde el programa de la CPU si hay que mostrar niebla en la distancia
uniform Spotlight spotlight;	// Luz de la linterna del jugador. Variable 'spotlight' que es una estructura de datos 'Spotlight'
uniform vec3 ambientLight;	// ADD 'ambientLight' llega desde la CPU con el color (RGB) de la luz ambiental del escenario 

// Declaración de la función 'CalculateSpotlight'
vec3 CalculateSpotlight(Spotlight light, vec3 normal, vec3 viewDir);

// =============================================================================
// Función inicial
// =============================================================================
void main()
{
	// Color y transparencia iniciales del pixel de la textura
	vec4 texColor = texture(meshTexture, vertexUv);

	// Hace que no se muestren los pixeles con transparencia por debajo de 0.1 (casi invisibles)
	if (texColor.a < 0.5f) 
		discard;	// 'discard' deja de procesar este código a partir de aquí

	// 0.7f para que las texturas oscuras no se vean muy oscuras sin la luz de la linterna
	float ambientFactor = 0.85f;	// 0.3f original
	//vec3 lightColor = vec3(0.5f, 0.972f, 0.905f);	// ABO
	vec3 lightColor = ambientLight;	// ADD
	vec3 ambient = ambientFactor * lightColor;

	vec3 norm = normalize(vertexNorms);
	vec3 lightDir = normalize(lightPos - fragPos);

	float diff = max(dot(norm, lightDir), 0.0f);
	vec3 diffuse = diff * lightColor;	 

	float specularFactor = 0.5f;
	vec3 viewDir = normalize(viewPos - fragPos);
	vec3 reflectDir = reflect(-lightDir, norm);
	float spec = pow(max(dot(viewDir, reflectDir), 0.0f), 32);
	vec3 specular = specularFactor * spec * lightColor;
	vec3 spotlightResult = CalculateSpotlight(spotlight, norm, viewDir);

	vec3 result; 	
	
	// Colorear el modelo 3D si ha recibido un impacto (solamente para enemigos)
	if (damaged || colorize_model != vec3(0.0f, 0.0f, 0.0f)) {
		// result = vec3(0.7f, 0.0f, 0.0f);	// OLD
		if (colorize_model.r != 0.0f) texColor.r = colorize_model.r;
		if (colorize_model.g != 0.0f) texColor.g = colorize_model.g;
		if (colorize_model.b != 0.0f) texColor.b = colorize_model.b;
	}

	// Calcular el color final del pixel con todos los factores que lo alteran:
	// Luz ambiental + difusión por 'normaMap' + brillo reflejado + luz linterna, multiplicado por el color original del pixel en la textura
	result = (ambient + diffuse + specular + spotlightResult) * vec3(texColor);	
    FragColor = vec4(result, 1.0f);
	
	// 'FragColor' contiene el color final y transparencia del pixel
	
	// Si existe niebla, 'FragColor' se oscurece en función de la distancia con respecto a la cámara del jugador.
	// Esta visibilidad se calcula en el código 'Vertex Shader' en la variable 'visibility'
	if (nightFog)
		FragColor = mix(vec4(0.0f, 0.0f, 0.0f, 1.0f), FragColor, visibility);
		
	// Al finalizar este código, el contenido de 'FragColor' es recogido automáticamente por el sistema de OpenGL que mezcla 
	// todos los colores de todos los pixeles (más otros procesos como filtros) en un buffer del 'frame' que dibuja en ese momento.
	
	// --- Fin del proceso 'Fragment Shader' ---
} 


// =============================================================================
// Definición de la función 'CalculateSpotlight'
// Calcula la incidencia de la luz de la linterna en las texturas con 'NormalMap'
// =============================================================================
vec3 CalculateSpotlight(Spotlight light, vec3 normal, vec3 viewDir)
{
	vec3 lightDir = normalize(light.position - fragPos);
	float diff = max(dot(normal, lightDir), 0.0f);
	
	vec3 reflectDirection = reflect(-lightDir, normal);
	float spec = pow(max(dot(viewDir, reflectDirection), 0), 32);
	
	float distance = length(light.position - fragPos);
	float attenuation = 1.0f / (light.constant + light.linear * distance + light.quadratic * (distance * distance));
	
	// Brillo de la luz
	float theta = dot(lightDir, normalize(-light.direction));
	float epsilon = light.cutOff - light.outerCutOff;
	float intensity = clamp((theta - light.outerCutOff) / epsilon, 0.0f, 1.0f);
	
	vec3 diffuse = light.diffuse;
	vec3 specular = light.specular;
	diffuse *= attenuation * intensity;
	specular *= attenuation * intensity;
	
	return (diffuse + specular);
}